import math
from pathlib import Path
from lightning import LightningDataModule

from torchchronos.datasets.ucr_uea_dataset import UCRUEADataset
from torchchronos.download import download_uea_ucr
from torch.utils.data import random_split, DataLoader

from ..utils import swap_batch_seq_collate_fn

from ..transforms import Compose, Transform


class UCRUEAModule(LightningDataModule):
    def __init__(
        self,
        name: str,
        split_ratio: tuple[float, float] = (0.75, 0.15),
        batch_size: int = 32,
        transform: Transform | Compose | None = None,
    ):
        super().__init__()

        self.batch_size = batch_size
        self.cache_dir = Path(".cache/data")
        self.split_ratio = split_ratio
        self.name = name
        self.transform = transform

    def prepare_data(self):
        download_uea_ucr(self.cache_dir, self.name)

    def label_from_index(self, index: int) -> str:
        assert self.__label_from_index is not None, "You need to call setup first"
        return self.__label_from_index(index)

    @property
    def num_classes(self) -> int:
        assert self.__num_classes is not None, "You need to call setup first"
        return self.__num_classes

    @property
    def dimensions(self) -> int:
        assert self.__dimensions is not None, "You need to call setup first"
        return self.__dimensions

    @property
    def series_length(self) -> int:
        assert self.__series_length is not None, "You need to call setup first"
        return self.__series_length

    @property
    def equal_length(self) -> bool:
        assert self.__equal_length is not None, "You need to call setup first"
        return self.__equal_length

    @property
    def univariate(self) -> bool:
        assert self.__univariate is not None, "You need to call setup first"
        return self.__univariate

    def setup(self, stage=None):
        dataset = UCRUEADataset(self.name, self.cache_dir, self.transform)
        self.__label_from_index = dataset.label_from_index
        self.__num_classes = dataset.num_classes
        self.__dimensions = dataset.dimensions
        self.__series_length = dataset.series_length
        self.__equal_length = dataset.equal_length
        self.__univariate = dataset.univariate

        # split dataset
        test_size = round(1 - self.split_ratio[0] - self.split_ratio[1], 2)
        if math.isclose(test_size, 0.0):
            self.train_dataset, self.val_dataset = random_split(
                dataset, [self.split_ratio[0], self.split_ratio[1]]
            )
        else:
            self.train_dataset, self.val_dataset, self.test_dataset = random_split(
                dataset, [self.split_ratio[0], self.split_ratio[1], test_size]
            )

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            collate_fn=swap_batch_seq_collate_fn,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            collate_fn=swap_batch_seq_collate_fn,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            collate_fn=swap_batch_seq_collate_fn,
        )



from pathlib import Path
import pandas as pd
from sktime.datasets import load_UCR_UEA_dataset
import torch
from torch.utils.data import Dataset

from ..transforms import Compose, Transform

from ..utils import get_project_root, parse_ts


class UCRUEADataset(Dataset):
    def __init__(
        self,
        ds_name: str,
        path: Path,
        transform: Transform | Compose | None = None,
    ) -> None:
        super().__init__()
        # It doesnt matter if test or train, but test is usually smaller
        ts_info = parse_ts(get_project_root() / path / ds_name / f"{ds_name}_TEST.ts")
        self.dimensions = ts_info.dimensions
        self.num_classes = ts_info.num_classes
        self.series_length = ts_info.series_length
        self.equal_length = ts_info.equal_length
        self.univariate = ts_info.univariate

        self.transform = transform
        self.xs, self.ys = load_UCR_UEA_dataset(
            ds_name,
            extract_path=get_project_root() / path,
            return_type="numpy3d",
        )
        self.xs = torch.tensor(self.xs, dtype=torch.float32).transpose(1, 2)
        if self.transform is not None:
            self.transform = self.transform.fit(self.xs)

        if self.ys.dtype == "U2" or self.ys.dtype == "<U1":
            # convert string labels to int
            factorized_y = pd.factorize(self.ys, sort=True)
            self.y_labels = factorized_y[1]
            self.ys = factorized_y[0]

        self.ys = torch.tensor(self.ys, dtype=torch.long)

    def label_from_index(self, index: int) -> str:
        return self.y_labels[int(index)]

    def __len__(self) -> int:
        return len(self.xs)

    def __getitem__(self, index: int) -> tuple[torch.FloatTensor, torch.LongTensor]:
        if self.transform:
            return self.transform(self.xs[index]), self.ys[index]
        return self.xs[index], self.ys[index]



from pathlib import Path
from typing import Optional
import zipfile

from sktime.datasets._data_io import _download_and_extract, _list_available_datasets


def download_uea_ucr(extract_path: Optional[Path], dataset_name: str):
    """This downloads the uea ucr dataset from sktime to the path extract_path."""
    # Download UCR/UEA archive from sktime
    if extract_path is not None:
        local_dirname = extract_path
    else:
        local_dirname = Path(".cache/data")
    local_dirname.mkdir(parents=True, exist_ok=True)

    if dataset_name not in _list_available_datasets(local_dirname):
        # Dataset is not already present in the datasets directory provided.
        # If it is not there, download and install it.
        url = "https://timeseriesclassification.com/Downloads/%s.zip" % dataset_name
        # This also tests the validitiy of the URL, can't rely on the html
        # status code as it always returns 200
        try:
            _download_and_extract(
                url,
                extract_path=local_dirname,
            )
        except zipfile.BadZipFile as e:
            raise ValueError(
                f"Invalid dataset name ={dataset_name} is not available on extract path ="
                f"{extract_path}. Nor is it available on "
                f"https://timeseriesclassification.com/.",
            ) from e
